#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QElapsedTimer>
#include <QDebug>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    ui->view->installEventFilter(this);

    InitGL(ui->view);
    Init();

    connect(&m_timer, SIGNAL(timeout()), this, SLOT(Draw()));
    m_timer.start(0);
}

MainWindow::~MainWindow()
{
    m_timer.stop();
    m_model.Release();
    m_SkyBox.Release();
    delete ui;
}

void MainWindow::InitGL(QWidget *pWidget)
{
    hWnd = (HWND)pWidget->winId();
    hDC = GetDC(hWnd);

    PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),                  // Size Of This Pixel Format Descriptor
        1,                              // Version Number
        PFD_DRAW_TO_WINDOW |                        // Format Must Support Window
        PFD_SUPPORT_OPENGL |                        // Format Must Support OpenGL
        PFD_DOUBLEBUFFER,                       // Must Support Double Buffering
        PFD_TYPE_RGBA,                          // Request An RGBA Format
        32,                               // Select Our Color Depth
        0, 0, 0, 0, 0, 0,                       // Color Bits Ignored
        0,                              // No Alpha Buffer
        0,                              // Shift Bit Ignored
        0,                              // No Accumulation Buffer
        0, 0, 0, 0,                         // Accumulation Bits Ignored
        32,                             // Z-Buffer (Depth Buffer)
        0,                              // Stencil Buffer
        0,                              // No Auxiliary Buffer
        PFD_MAIN_PLANE,                         // Main Drawing Layer
        0,                              // Reserved
        0, 0, 0                             // Layer Masks Ignored
    };

    int pf = ChoosePixelFormat(hDC, &pfd);
    SetPixelFormat(hDC, pf, &pfd);
    hRC = wglCreateContext(hDC);
    wglMakeCurrent(hDC, hRC);

    glewInit();
}

void MainWindow::Init()
{
    wglMakeCurrent(hDC, hRC);

    // openal
    alutInit(&__argc, __argv);
    ALuint buffer;
    buffer = alutCreateBufferFromFile("Sound.wav");

    ALuint source;
    alGenSources(1, &source);
    alSourcei(source, AL_BUFFER, buffer);
    alSourcei(source, AL_LOOPING, true);
    alSourcePlay(source);

    // opengl
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    //glEnable(GL_CULL_FACE);
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPatchParameteri(GL_PATCH_VERTICES, 3);
    wglSwapIntervalEXT(0);

    m_shaderDraw.Load("Draw.vs.txt", "Draw.fs.txt", "Draw.cs.txt", "Draw.es.txt");
    m_shaderShadowMap.Load("DrawToDepthTexture.vs.txt", "DrawToDepthTexture.fs.txt");
    std::vector<int> listTypes;
    listTypes.push_back(GL_RGBA32F);
    m_RenderToShadowTexture.Load(listTypes, m_nShadowWidth, m_nShadowWidth);

    m_model.Load("OBJ", "Scene.obj");
    m_model.CreateOpenGLBuffers();
    m_model.SetDispFactor(0.05f);

    m_camera.Init(glm::vec3(0,8,20), glm::vec3(0,0,0));
    for(int i = 0; i < 256; i++)
    {
        SetKeys(i, false);
    }

    m_SkyBox.Load("OBJ/", "barren_bk.jpg", "barren_dn.jpg", "barren_ft.jpg", "barren_lf.jpg", "barren_rt.jpg", "barren_up.jpg");

    m_elapsedTimer.start();
    m_nElapsedTime = m_nCurrentTime = m_elapsedTimer.nsecsElapsed();
}

void MainWindow::Draw()
{
    wglMakeCurrent(hDC, hRC);

    // getDT
    m_nElapsedTime = m_nCurrentTime;
    m_nCurrentTime = m_elapsedTimer.nsecsElapsed();
    dt = (float)(m_nCurrentTime - m_nElapsedTime) / 1000000000.0f;

    // print fps
    nFPS++;
    fSec += dt;
    if (fSec >= 1.0f)
    {
        this->setWindowTitle("FPS: " + QString::number(nFPS));

        nFPS = 0;
        fSec = 0.0f;
    }

    int nWidth = ui->view->width();
    int nHeight = ui->view->height();
    if (nWidth < 1) { nWidth = 1; }
    if (nHeight < 1) { nHeight = 1; }

    // update
    if (true == bKeys[Qt::Key_W]) { m_camera.MoveW(); }
    if (true == bKeys[Qt::Key_S]) { m_camera.MoveS(); }
    if (true == bKeys[Qt::Key_D]) { m_camera.MoveD(); }
    if (true == bKeys[Qt::Key_A]) { m_camera.MoveA(); }
    m_camera.Update(dt);

    glm::vec3 v3CameraPos = m_camera.GetPos();
    glm::vec3 v3CameraAt = m_camera.GetAt();
    glm::vec3 v3CameraDir = glm::normalize(v3CameraAt - v3CameraPos);

    glm::vec3 v3LightPos = glm::vec3(32.6785f, 85.7038f, -39.8369f);
    glm::vec3 v3LightAt = glm::vec3(0, 0, 0);
    glm::vec3 v3LightDir = glm::normalize(v3LightAt - v3LightPos);

    glm::mat4 mLightWorld = glm::mat4(1.0f);
    glm::mat4 mLightView = glm::lookAtRH(v3LightPos, v3LightAt, glm::vec3(0, 1, 0));
    glm::mat4 mLightProj = glm::orthoRH(-30.0f,30.0f, -30.0f,30.0f, 1.0f, 200.0f);


    // draw
    // 1/2 - draw to shadow texture
    m_RenderToShadowTexture.Bind();
    //
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glViewport(0, 0, m_nShadowWidth, m_nShadowWidth);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //
    m_shaderShadowMap.Begin();
    glm::mat4 mWorld = glm::mat4(1.0f);
    m_shaderShadowMap.SetMatrix("matWorld", &mWorld);
    m_shaderShadowMap.SetMatrix("matView", &mLightView);
    m_shaderShadowMap.SetMatrix("matProj", &mLightProj);

    m_model.Draw(&m_shaderShadowMap, GL_TRIANGLES);

    m_shaderShadowMap.End();
    m_RenderToShadowTexture.Unbind();



    // draw screen to texture
    glm::mat4 mCameraWorld = glm::mat4(1.0f);
    glm::mat4 mCameraView = glm::lookAtRH(v3CameraPos, v3CameraAt, glm::vec3(0, 1, 0));
    glm::mat4 mCameraProj = glm::perspectiveRH(glm::radians(45.0f), (float)nWidth / (float)nHeight, 0.1f, 600.0f);

    glClearColor(0.5f, 0.5f, 1.0f, 1.0f);
    glViewport(0, 0, nWidth, nHeight);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(glm::value_ptr(mCameraProj));
    glMatrixMode(GL_MODELVIEW);
    mWorld = glm::mat4(1.0f);
    glLoadMatrixf(glm::value_ptr(mCameraView * mWorld));

    m_shaderDraw.Begin();

    mWorld = glm::mat4(1.0f);
    glm::mat4 mVP = mCameraProj * mCameraView;
    m_shaderDraw.SetMatrix("gWorld", &mWorld);
    m_shaderDraw.SetVector3("gEyeWorldPos", &v3CameraPos);
    m_shaderDraw.SetMatrix("gVP", &mVP);
    m_shaderDraw.SetVector3("lightDir", &v3LightDir);

    m_shaderDraw.SetMatrix("matLightView", &mLightView);
    m_shaderDraw.SetMatrix("matLightProj", &mLightProj);
    m_shaderDraw.SetTexture("g_DepthTexture", m_RenderToShadowTexture.GetTextureID(0), 2);

    if (true == m_bIsDrawWireframe)
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    }
    else
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    m_model.Draw(&m_shaderDraw, GL_PATCHES);

    m_shaderDraw.DisableTexture(2);
    m_shaderDraw.End();

    m_SkyBox.Draw(v3CameraPos, 300.0f);

    SwapBuffers(hDC);
}

bool MainWindow::eventFilter(QObject *obj, QEvent *event) {
    bool ret = QObject::eventFilter(obj, event);

    if (event->type() == QEvent::Resize)
    {
        ;
    }

    return ret;
}

void MainWindow::keyPressEvent(QKeyEvent *event)
{
    SetKeys(event->key(), true);

    if (Qt::Key_Escape == event->key())
    {
        close();
    }

    if (Qt::Key_Space == event->key())
    {
        m_bIsDrawWireframe = !m_bIsDrawWireframe;
    }
}

void MainWindow::keyReleaseEvent(QKeyEvent *event)
{
    SetKeys(event->key(), false);
}

void MainWindow::mousePressEvent(QMouseEvent *event)
{
    m_bMouseButtonDown = true;
}

void MainWindow::mouseReleaseEvent(QMouseEvent *event)
{
    m_bMouseButtonDown = false;
}

void MainWindow::mouseMoveEvent(QMouseEvent *event)
{
    if (true == m_bMouseButtonDown)
    {
        m_camera.Rotate(event->pos().x(), event->pos().y());
    }
}

void MainWindow::SetKeys(int id, bool bPressed)
{
    if (id >= 0 && id <= 255)
    {
        bKeys[id] = bPressed;
    }
}
