#include "Model.h"

#include "assimp/Importer.hpp"
#include "assimp/postprocess.h"
#include "assimp/scene.h"

#include <QFileInfo>

Model::Model()
{

}

Model::~Model()
{
    Release();
}

void Model::Load(std::string strDir, std::string strFilename, glm::mat4 matTransform, bool bIsLoadTextures)
{
    Assimp::Importer importer;
    const aiScene *pScene = importer.ReadFile((strDir + "/" + strFilename), aiProcessPreset_TargetRealtime_MaxQuality);

    m_bIsLoadTextures = bIsLoadTextures;
    for (unsigned int m = 0; m < pScene->mNumMaterials; m++)
    {
        Material material;

        if (true == bIsLoadTextures)
        {
            if (pScene->mMaterials[m]->GetTextureCount(aiTextureType_DIFFUSE) > 0)
            {
                aiString path;
                pScene->mMaterials[m]->GetTexture(aiTextureType_DIFFUSE, 0, &path);
                std::string strTextureFilename(path.C_Str());

                material.m_pAlbedoTexture = new Texture();
                material.m_pAlbedoTexture->Load(strDir + "/" + strTextureFilename);

                QFileInfo fileInfo(QString((strDir + "/" + strTextureFilename).c_str()));
                QString name = fileInfo.completeBaseName();
                QString ext = fileInfo.suffix();
                material.m_pDispTexture = new Texture();

                QFileInfo fileInfoDisp(QString(strDir.c_str()) + "/" + name + "_DISP." + ext);
                if (true == fileInfoDisp.exists())
                {
                    material.m_pDispTexture->Load(strDir + "/" + name.toStdString() + "_DISP." + ext.toStdString());
                }
                else
                {
                    material.m_pDispTexture->Load(strDir + "/" + "none_DISP.jpg");
                }
            }
        }

        m_listMaterials.push_back(material);
    }

    for(unsigned int m = 0; m < pScene->mNumMeshes; m++)
    {
        aiMesh *pMesh = pScene->mMeshes[m];

        Material &material = m_listMaterials[pMesh->mMaterialIndex];

        for(unsigned int f = 0; f < pMesh->mNumFaces; f++)
        {
            aiFace face = pMesh->mFaces[f];

            for(int i = 0; i < 3; i++)
            {
                aiVector3D v = pMesh->mVertices[ face.mIndices[i] ];
                aiVector3D n = pMesh->mNormals[ face.mIndices[i] ];
                aiVector3D t = pMesh->mTextureCoords[0][ face.mIndices[i] ];

                Vertex vertex;
                vertex.v3Position = glm::vec3(matTransform * glm::vec4(v.x, v.y, v.z, 1.0f));
                vertex.v3Normal = glm::vec3(matTransform * glm::vec4(n.x, n.y, n.z, 0.0f));
                vertex.v3Normal = glm::normalize(vertex.v3Normal);
                vertex.v2TextCoord = glm::vec2(t.x, t.y);

                material.m_listVertices.push_back(vertex);
            }
        }
    }

    ApplyTessellation(2.0f);


    int id = 0;
    for (int m = 0; m < (int)m_listMaterials.size(); m++)
    {
        Material& material = m_listMaterials[m];

        for (int i = 0; i < (int)material.m_listVertices.size(); i++)
        {
            Vertex vertex = material.m_listVertices[i];
            m_listAllVertices.push_back(vertex);

            material.m_listIndices.push_back(id);
            id++;
        }
    }
}

float Model::GetTriangleArea(glm::vec3 A, glm::vec3 B, glm::vec3 C)
{
    float fArea = glm::length( glm::cross(B - A, C - A) ) / 2.0f;
    return fArea;
}

void Model::ApplyTessellation(float fLength)
{
    for(uint m = 0; m < m_listMaterials.size(); m++)
    {
        Material &material = m_listMaterials[m];

        // tessenlation
        QList< Vertex > tessenlationVertices;
        for(uint i = 0; i < material.m_listVertices.size(); i += 3)
        {
            Vertex A = material.m_listVertices[i + 0];
            Vertex B = material.m_listVertices[i + 1];
            Vertex C = material.m_listVertices[i + 2];

            Tessellation(A, B, C, fLength, tessenlationVertices);
        }

        // copy
        material.m_listVertices.clear();
        for(int i = 0; i < tessenlationVertices.size(); i++)
        {
            material.m_listVertices.push_back(tessenlationVertices[i]);
        }
        tessenlationVertices.clear();
    }
}

void Model::Tessellation(Vertex A, Vertex B, Vertex C, float fLength, QList< Vertex > &tessellationVertices)
{
    float dist0 = glm::distance(A.v3Position, B.v3Position);
    float dist1 = glm::distance(B.v3Position, C.v3Position);
    float dist2 = glm::distance(C.v3Position, A.v3Position);
    float fCurrentLength = fmax(dist0, fmax(dist1, dist2));

    if (fLength < fCurrentLength)
    {
        Vertex D; GetCenter(D, A, B);
        Vertex E; GetCenter(E, B, C);
        Vertex F; GetCenter(F, C, A);

        Tessellation(A, D, F, fLength, tessellationVertices);
        Tessellation(D, E, F, fLength, tessellationVertices);
        Tessellation(D, B, E, fLength, tessellationVertices);
        Tessellation(F, E, C, fLength, tessellationVertices);
    }
    else
    {
        tessellationVertices.push_back(A);
        tessellationVertices.push_back(B);
        tessellationVertices.push_back(C);
    }
}

void Model::GetCenter(Vertex &ret, Vertex A, Vertex B)
{
    ret.v3Position = (A.v3Position * 0.5f) + (B.v3Position * 0.5f);
    ret.v3Normal = (A.v3Normal * 0.5f) + (B.v3Normal * 0.5f);
    ret.v2TextCoord = (A.v2TextCoord * 0.5f) + (B.v2TextCoord * 0.5f);

    ret.v3Normal = glm::normalize(ret.v3Normal);
}


void Model::CreateOpenGLBuffers()
{
    glGenBuffers(1, &m_glVertexBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, m_glVertexBuffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * m_listAllVertices.size(), m_listAllVertices.data(), GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    for(int m = 0; m < m_listMaterials.size(); m++)
    {
        Material &material = m_listMaterials[m];

        if (material.m_listIndices.size() == 0)
        {
            continue;
        }

        glGenBuffers(1, &material.m_glIndexBuffer);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, material.m_glIndexBuffer);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint32_t) * material.m_listIndices.size(), material.m_listIndices.data(), GL_STATIC_DRAW);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}

void Model::Draw(Shader *shader, uint32_t mode)
{
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glEnableVertexAttribArray(2);
    glBindBuffer(GL_ARRAY_BUFFER, m_glVertexBuffer);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)0);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)(sizeof(glm::vec3)) );
    glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)(sizeof(glm::vec3) + sizeof(glm::vec3)) );

    for(unsigned int m = 0; m < m_listMaterials.size(); m++)
    {
        Material &material = m_listMaterials[m];

        if (material.m_listVertices.size() == 0)
        {
            continue;
        }

        if (true == m_bIsLoadTextures)
        {
            if (nullptr == material.m_pAlbedoTexture)
            {
                continue;
            }

            if (nullptr == material.m_pDispTexture)
            {
                continue;
            }

            shader->SetTexture("g_Texture", material.m_pAlbedoTexture, 0);
            shader->SetTexture("g_DispTexture", material.m_pDispTexture, 1);
            shader->SetFloat("gDispFactor", m_fDispFactor);
        }

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, material.m_glIndexBuffer);
        glDrawElements(mode, (int)material.m_listIndices.size(), GL_UNSIGNED_INT, nullptr);
    }

    shader->DisableTexture(0);
    shader->DisableTexture(1);

    glDisableVertexAttribArray(0);
    glDisableVertexAttribArray(1);
    glDisableVertexAttribArray(2);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    //glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void Model::Release()
{
    for(int m = 0; m < m_listMaterials.size(); m++)
    {
        Material &material = m_listMaterials[m];

        material.m_listVertices.clear();

        if (nullptr != material.m_pAlbedoTexture)
        {
            material.m_pAlbedoTexture->Release();
            delete material.m_pAlbedoTexture;
            material.m_pAlbedoTexture = nullptr;
        }

        if (nullptr != material.m_pDispTexture)
        {
            material.m_pDispTexture->Release();
            delete material.m_pDispTexture;
            material.m_pDispTexture = nullptr;
        }
    }

    m_listMaterials.clear();
    m_listAllVertices.clear();
}

std::vector< Vertex >* Model::GetVertices()
{
    return &m_listAllVertices;
}

std::vector< Material >* Model::GetMaterials()
{
    return &m_listMaterials;
}

void Model::SetDispFactor(float fValue)
{
    m_fDispFactor = fValue;
}

float Model::GetDispFactor()
{
    return m_fDispFactor;
}
