#pragma once


#include "GL/glew.h"
#include "GL/wglew.h"
#include "glm/glm.hpp"
#include "glm/ext.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtx/transform2.hpp"
#include "texture.h"

#include <QDebug>

#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>

class Shader
{
    GLuint ProgramID;
public:
    bool Load(QString strVertexShaderFilename, QString strFragmentShaderFilename, QString strTessControlShader = nullptr, QString strTessEvaluationShader = nullptr)
	{
        std::vector<GLchar> sourcecode;

        LoadFileContents(strVertexShaderFilename.toStdString(), sourcecode);
        GLuint vertex_shader = CompileShader(sourcecode, GL_VERTEX_SHADER);

        GLuint tess_control_shader = 0;
        if (nullptr != strTessControlShader && "" != strTessControlShader)
        {
            LoadFileContents(strTessControlShader.toStdString(), sourcecode);
            tess_control_shader = CompileShader(sourcecode, GL_TESS_CONTROL_SHADER);
        }

        GLuint tess_evaluation_shader = 0;
        if (nullptr != strTessEvaluationShader && "" != strTessEvaluationShader)
        {
            LoadFileContents(strTessEvaluationShader.toStdString(), sourcecode);
            tess_evaluation_shader = CompileShader(sourcecode, GL_TESS_EVALUATION_SHADER);
        }

        LoadFileContents(strFragmentShaderFilename.toStdString(), sourcecode);
        GLuint frag_shader = CompileShader(sourcecode, GL_FRAGMENT_SHADER);

        ProgramID = glCreateProgram();

        glAttachShader(ProgramID, vertex_shader);
        if (0 != tess_control_shader)
        {
            glAttachShader(ProgramID, tess_control_shader);
        }
        if (0 != tess_evaluation_shader)
        {
            glAttachShader(ProgramID, tess_evaluation_shader);
        }
        glAttachShader(ProgramID, frag_shader);

        glDeleteShader(vertex_shader);
        if (0 != tess_control_shader)
        {
            glDeleteShader(tess_control_shader);
        }
        if (0 != tess_evaluation_shader)
        {
            glDeleteShader(tess_evaluation_shader);
        }
        glDeleteShader(frag_shader);

        glLinkProgram(ProgramID);

        GLint result = GL_TRUE;
        glGetProgramiv(ProgramID, GL_LINK_STATUS, &result);

        if(result == GL_FALSE)
        {
            GLint length = 0;
            std::vector<char> log;

            glGetProgramiv(ProgramID, GL_INFO_LOG_LENGTH, &length);

            log.resize(length);

            glGetProgramInfoLog(ProgramID, length, &result, &log[0]);

            glDeleteProgram(ProgramID);

            qDebug() << QString( std::string(log.begin(), log.end()).c_str() );
            throw std::string(log.begin(), log.end());
        }

        return true;
	}

    GLuint CompileShader(std::vector<GLchar> const& shader_source, GLenum type)
    {
        GLuint shader = glCreateShader(type);

        GLint len = static_cast<GLint>(shader_source.size());
        GLchar const* source_array = &shader_source[0];

        glShaderSource(shader, 1, &source_array, &len);
        glCompileShader(shader);

        GLint result = GL_TRUE;
        glGetShaderiv(shader, GL_COMPILE_STATUS, &result);

        if(result == GL_FALSE)
        {
            std::vector<char> log;

            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &len);

            log.resize(len);

            glGetShaderInfoLog(shader, len, &result, &log[0]);

            glDeleteShader(shader);

            qDebug() << QString( std::string(log.begin(), log.end()).c_str() );
            throw std::string(log.begin(), log.end());

            return 0;
        }

        return shader;
    }

    void LoadFileContents(std::string const& name, std::vector<char>& contents, bool binary = false)
    {
        std::ifstream in(name, std::ios::in | (std::ios_base::openmode)(binary?std::ios::binary : 0));

        if (in)
        {
            contents.clear();

            std::streamoff beg = in.tellg();

            in.seekg(0, std::ios::end);

            std::streamoff fileSize = in.tellg() - beg;

            in.seekg(0, std::ios::beg);

            contents.resize(static_cast<unsigned>(fileSize));

            in.read(&contents[0], fileSize);
        }
        else
        {
            throw std::string("Cannot read the contents of a file");
        }
    }

	void Release()
	{
		glDeleteProgram(ProgramID);
	}

	void Begin()
	{
		glUseProgram(ProgramID);
	}
	void End()
	{
		glUseProgram(0);
	}

    void SetFloat(const char *varname, float val)
    {
        GLint loc = glGetUniformLocation(ProgramID, varname);
        glUniform1f(loc, val);
    }

    void SetVector3(const char *varname, glm::vec3 *val)
    {
        GLint loc = glGetUniformLocation(ProgramID, varname);
        glUniform3fv(loc, 1, (float*)&(val[0]));
    }

    void SetVector4(const char *varname, glm::vec4 *val)
    {
        GLint loc = glGetUniformLocation(ProgramID, varname);
        glUniform4fv(loc, 1, (float*)&(val[0]));
    }

    void SetVector4(const char *varname, glm::vec3 *val, float w)
    {
        glm::vec4 val2(*val, w);

        GLint loc = glGetUniformLocation(ProgramID, varname);
        glUniform4fv(loc, 1, (float*)&(val2[0]));
    }

    void SetMatrix(const char *varname, glm::mat4 *val)
	{
		GLint loc = glGetUniformLocation(ProgramID, varname);
		glUniformMatrix4fv(loc, 1, GL_FALSE, (float*)&(val[0][0]));
	}
    void SetMatrixArray(const char *varname, glm::mat4 *val, unsigned int count)
	{
		GLint loc = glGetUniformLocation(ProgramID, varname);
		glUniformMatrix4fv(loc, count, GL_FALSE, (float*)&(val[0][0]));
	}
	void SetTexture(const char *varname, Texture *texture, unsigned int i)
	{
		glActiveTexture(GL_TEXTURE0 + i);
		glBindTexture(GL_TEXTURE_2D, texture->ID());

		GLint loc = glGetUniformLocation(ProgramID, varname);
		glUniform1i(loc, i);
	}
    void SetTexture(const char *varname, GLuint id, unsigned int i)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_2D, id);

        GLint loc = glGetUniformLocation(ProgramID, varname);
        glUniform1i(loc, i);
    }
    void DisableTexture(unsigned int i)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_2D, 0);
    }

	void BindVector4Buffer(const char *attribName, GLuint buffer)
	{
		GLuint attribID = glGetAttribLocation(ProgramID, attribName);

		glEnableVertexAttribArray(attribID);
		glBindBuffer(GL_ARRAY_BUFFER, buffer);
		glVertexAttribPointer
			(
			attribID,                     // The attribute we want to configure
			4,                            // size
			GL_FLOAT,                     // type
			GL_FALSE,                     // normalized?
			0,                            // stride
			(void*)0                      // array buffer offset
			);
	}
	void BindVector3Buffer(const char *attribName, GLuint buffer)
	{
		GLuint attribID = glGetAttribLocation(ProgramID, attribName);

		glEnableVertexAttribArray(attribID);
		glBindBuffer(GL_ARRAY_BUFFER, buffer);
		glVertexAttribPointer
			(
			attribID,                     // The attribute we want to configure
			3,                            // size : U+V => 2
			GL_FLOAT,                     // type
			GL_FALSE,                     // normalized?
			0,                            // stride
			(void*)0                      // array buffer offset
			);
	}
	void BindVector2Buffer(const char *attribName, GLuint buffer)
	{
		GLuint attribID = glGetAttribLocation(ProgramID, attribName);

		glEnableVertexAttribArray(attribID);
		glBindBuffer(GL_ARRAY_BUFFER, buffer);
		glVertexAttribPointer
			(
			attribID,                     // The attribute we want to configure
			2,                            // size : U+V => 2
			GL_FLOAT,                     // type
			GL_FALSE,                     // normalized?
			0,                            // stride
			(void*)0                      // array buffer offset
			);
	}
	void UnbindBuffer(const char *attribName)
	{
		GLuint attribID = glGetAttribLocation(ProgramID, attribName);
		glDisableVertexAttribArray(attribID);

	}
	void DrawTriangles(unsigned int first, unsigned int count)
	{
		glDrawArrays(GL_TRIANGLES, first, count);
	}
};
